/*
 ============================================================================
 multikeys - an advanced hotkeys module for LiteStep/PureLS
 ============================================================================

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 ============================================================================

multikeys 3.1 (2001-12-22) / qwilk 
	- Added LsBox support (mkLsBoxEnable, !mkLsBoxHook), which enables
	  dragging, alpha transparency etc!

multikeys 3.0 (2001-11-12) / qwilk 
	- Added support for jKey and standard LS Hotkey syntax! (you can disable
	  this by using mkNoLoadjKeys and mkNoLoadHotkeys, respectively) 
	- Changed the name of the module to multikeys to reflect its support for
	  multiple groups, multiple syntaxes, and multiple shells (LS/PureLS) 
	- All bangs now begin with "!mk", all step.rc settings with "mk",
	  and key definitions with "*mkey" (should be easier to remember <vbg>) 
	- Default group is now set to "Default" if not defined, unless jKeys/Hotkeys
	  are present, then "jKey"/"Hotkey" will be used instead (jKey has precedence
	  if you use both...) 
	- Added !mkPlayWav <.wav> (only 10 more lines of code - one less module to load!) 
	- Added missing key: PrtScn 
	- Fixed: Error color would sometimes be used for successful commands when
	  toggling the window from hidden to shown 
	- Fixed: Settings can now be in any file (PureLS fix, thanks jugg for the tip!) 
	- Changed group switching bang (again...) to "!mkGroup <group>" (the logical choice) 
	- New documentation (HTML) 
	- Cleaned up the source code 

 ============================================================================

jukeys 2.0q (2001-02-07) / qwilk
	- Added jukeysOffsetX <offset>, jukeysOffsetY <offset>
	  (works the same way as in LSXCommand)
	- Added jukeysGroupShow, jukeysGroupX <offset>, jukeysGroupY <offset>
	- Removed jukeysNoTitle (replaced by the new Group commands)
	- Added jukeysCommandShow, jukeysCommandX <offset>, jukeysCommandY <offset>
	- Added jukeysCommandColor <color>, jukeysErrorColor <color>
	- Added jukeysMsgErrorCmd <message>, jukeysMsgErrorBang <message>
	- Added jukeysErrorSound <path\to\.wav>
	- Added jukeysMouseLeft <n>, jukeysMouseRight <n>, jukeysMouseMiddle <n>
	- Added a new "select group" popup menu for easy group switching
	- Added !jukeysPopup, !jukeysPopupGroups
	- Added !jukeysRepeat
	- Added !jukeysMove, !jukeysMove <x> <y>
	- Added !jukeysSetCommand <command> <args> and !jukeysSetError <message>
	  as an interface to the "world outside" (other modules)... =)
	- Added missing key: Escape
	- Now repositions automatically on screen resolution change
	- Change: "!jkSwitchTo <group>" is now "!jukeysSwitchTo <group>"
	- Change: A background bitmap now has precedence over background colour
	- Fixed: Everything (?) is updated on !Refresh now
	- Fixed: Multi-modifier hotkeys wasn't displayed correctly in popup menu
	- Fixed: jukeysFontFace wasn't working
	- Fixed: Switching to a non-existing group was possible
	- Fixed: Bang scripting (!execute [command1][command2]... ) now works
	- Included a background bitmap, an error sound and a jukeys.rc configuration file
	- Cleaned up the code, added more comments (source to be released soon)

	jukeys 1.1q (2001-01-22) / qwilk
	- Added jukeysNoTitle (-> current group name not displayed)
	- Added missing keys: Pause, Insert, Delete, Home, End,
	  PageUp, PageDown, Tab, Backspace, Spacebar, Apps, Enter,
	  Mul, Div, Add, Sub, Dec, Num0-Num9
	- !Refresh support (only hotkey definitions are refreshed for now)
	- Compiled with latest LSAPI; now uses LSExecuteEx instead

	jukeys 1.0 (2000-09-26) / MrJukes
	- Original release

 ============================================================================ */

#include "multikeys.h"

//#ifndef _DEBUG
#pragma warning ( disable : 4100 )
//#endif

//===========================================================================

char* szAppName = "multikeys";
char* szVersion = "multikeys 3.1 (qwilk)";
char* szDetailed = "multikeys.dll version 3.1 date 2001-12-22 (qwilk)";

//===========================================================================
// Module initialization and cleanup
//===========================================================================

int initModuleEx(HWND parent, HINSTANCE dllInst, LPCSTR szPath)
{
	WNDCLASS wc;
	strcpy(LSdir, szPath);
	hwndParent = parent;
	hInstance = dllInst;
	memset(&wc,0,sizeof(wc));
	wc.lpfnWndProc = WndProc;       // our window procedure
	wc.hInstance = dllInst;         // hInstance of DLL
	wc.lpszClassName = szAppName;   // our window class name
	
	if (!RegisterClass(&wc)) 
	{
		MessageBox(parent, "Error: Could not register window class", szAppName, MB_OK | MB_ICONERROR);
		return 1;
	}

	multikeysLoadSetup();

	hwndMain = CreateWindowEx(WS_EX_TOOLWINDOW, szAppName, szAppName, WS_POPUP | WS_CLIPCHILDREN | WS_CLIPSIBLINGS, x, y, w, h, NULL, NULL, dllInst, NULL);
	if (!hwndMain) return 1;
	SetWindowLong(hwndMain, GWL_USERDATA, magicDWord);
	SendMessage(hwndParent, LM_REGISTERMESSAGE, (WPARAM)hwndMain, (LPARAM)msgs);

	if (bgbitmap && !UseBgColor) SetWindowBitmapRgn(hwndMain, bgbitmap);
	mkStartHidden = (GetRCBool("mkStartHidden", true) ? true:false);
	mkLsBoxEnable = (GetRCBool("mkLsBoxEnable", true) ? true:false);
	if (!mkStartHidden && !mkLsBoxEnable) ShowWindow(hwndMain, SW_SHOW);

	//Hide window if neither bgbitmap or bgcolor is set (for plug'n'play jKey/Hotkey compatibility)
	if (!UseBgBitmap && !UseBgColor) ShowWindow(hwndMain, SW_HIDE);
	
	mkAlwaysOnTop = (GetRCBool("mkAlwaysOnTop", true) ? true:false);
	if (mkAlwaysOnTop) SetWindowPos(hwndMain, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOACTIVATE | SWP_NOSIZE | SWP_NOMOVE);

	multikeysAddBangs();
	LoadGroupHotkeys();
	SwitchToGroup(default_group);

	return 0;
}

//--------------------

void quitModule(HINSTANCE dll)
{
	SendMessage(hwndParent, LM_UNREGISTERMESSAGE, (WPARAM)hwndMain, (LPARAM)msgs);
	multikeysDestroyHotkeys();
	multikeysRemoveBangs();
	DestroyWindow(hwndMain);
	DeleteObject(bgbrush);
	DeleteObject(bgbitmap);
	DeleteObject(font);
	UnregisterClass(szAppName, dll);
}

//===========================================================================
// !Bang commands
//===========================================================================

void BangShow(HWND caller, const char* args)
{
	if (!(mkLsBoxEnable && !HookedToLsBox)) { ShowWindow(hwndMain, SW_SHOW); }
}
void BangHide(HWND caller, const char* args)
{
	if (!(mkLsBoxEnable && !HookedToLsBox)) { ShowWindow(hwndMain, SW_HIDE); }
}
void BangToggle(HWND caller, const char* args)
{
	if (!(mkLsBoxEnable && !HookedToLsBox))
	{
		if (IsWindowVisible(hwndMain)) BangHide(caller, args);
		else BangShow(caller, args);
	}
}

void BangPopup(HWND caller, const char* args) { multikeysPopupMenu(); }
void BangPopupGroups(HWND caller, const char* args) { multikeysPopupGroups(); }
void BangGroup(HWND caller, const char* args) { SwitchToGroup(args); }
void BangRepeat(HWND caller, const char* args) { multikeysRepeatCommand(); }
void BangMove(HWND caller, const char *args) { multikeysMove(caller, args); }
void BangSetCommand(HWND caller, const char *args) { multikeysSetCommand(caller, args); }
void BangSetError(HWND caller, const char *args) { multikeysSetError(caller, args); }
void BangPlayWav(HWND caller, const char *args)
{
	if(args && *args)
	{
		sprintf(mkPlayWav, "%s", args);
		PlaySound(mkPlayWav, NULL, SND_FILENAME | SND_NOWAIT);
		//adding | SND_ASYNC plays using asynchronous sound, but it may skip from time to time
	}
}

void BangBoxHook(HWND caller, const char *args)
{
	char *handle = strrchr(args,' ');
	if (handle && mkLsBoxEnable) 
	{
		HWND hwndBox = (HWND)atoi(handle+1);
		if (hwndBox) 
		{
			HookedToLsBox = true;
			if (hwndBox != hwndParent)
			{
				SetWindowLong(hwndMain, GWL_STYLE, (GetWindowLong(hwndMain, GWL_STYLE) &~ WS_POPUP)|WS_CHILD);
				SetParent(hwndMain, hwndBox);
				// This is a *bad* idea under Win2k SP2 - don't ask me why... [BlkHawk's comment]
				// SetWindowLong(hwndMain, GWL_STYLE, (GetWindowLong(hwndMain, GWL_STYLE) &~ WS_CHILD)|WS_POPUP);
				if (!mkStartHidden) { ShowWindow(hwndMain, SW_SHOWNORMAL); }
			}
		}
	}
return;
}


//===========================================================================
// !Bang command handling (called by initModuleEx & quitModule)
//===========================================================================

void multikeysAddBangs()
{
	AddBangCommand("!mkShow", BangShow);
	AddBangCommand("!mkHide", BangHide);
	AddBangCommand("!mkToggle", BangToggle);
	AddBangCommand("!mkGroup", BangGroup);
	AddBangCommand("!mkPopup", BangPopup);
	AddBangCommand("!mkPopupGroups", BangPopupGroups);
	AddBangCommand("!mkRepeat", BangRepeat);
	AddBangCommand("!mkMove", BangMove);
	AddBangCommand("!mkSetCommand", BangSetCommand);
	AddBangCommand("!mkSetError", BangSetError);
	AddBangCommand("!mkPlayWav", BangPlayWav);
	AddBangCommand("!mkLsBoxHook", BangBoxHook);
}

//--------------------

void multikeysRemoveBangs()
{
	RemoveBangCommand("!mkShow");
	RemoveBangCommand("!mkHide");
	RemoveBangCommand("!mkToggle");
	RemoveBangCommand("!mkGroup");
	RemoveBangCommand("!mkPopup");
	RemoveBangCommand("!mkPopupGroups");
	RemoveBangCommand("!mkRepeat");
	RemoveBangCommand("!mkMove");
	RemoveBangCommand("!mkSetCommand");
	RemoveBangCommand("!mkSetError");
	RemoveBangCommand("!mkPlayWav");
	RemoveBangCommand("!mkLsBoxHook");
}

//===========================================================================
// Module settings
//===========================================================================

void multikeysLoadSetup()
{
	int size;

	ScreenX = GetSystemMetrics(SM_CXSCREEN);
	ScreenY = GetSystemMetrics(SM_CYSCREEN);

	x = GetRCInt("mkX", 0); rcx = x;
	y = GetRCInt("mkY", 0); rcy = y;
	w = GetRCInt("mkW", 400);
	h = GetRCInt("mkH", 25);

	//	"Reference from" offset as used by LSXCommand:
	//	0 -> Left/Top (default), 1 -> Horizontal/Vertical Center, 2 -> Bottom/Right
	mkOffsetX = GetRCInt("mkOffsetX", 0);
	mkOffsetY = GetRCInt("mkOffsetY", 0);

	//recalculate relative coordinates to screen coordinates
	multikeysGetPosition();

	GetRCString("mkFontFace", temp, "Arial", 256);
	size = GetRCInt("mkFontSize", 14);
	font = CreateFont(size, 0, 0, 0, FW_DONTCARE, false, false, false, ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH, temp);
	color = GetRCColor("mkFontColor", 0x00FFFFFF);
	cmdcolor = GetRCColor("mkCommandColor", color);
	errorcolor = GetRCColor("mkErrorColor", cmdcolor);

	//Background bitmap has precedence over background color!
	UseBgBitmap = (GetRCBool("mkBackBmp", true) ? true:false);
	GetRCString("mkBackBmp", temp, "", 256);
	bgbitmap = LoadLSImage(temp, temp);

	UseBgColor = (GetRCBool("mkBackColor", true) ? true:false);
	if (UseBgColor) 
	{
		COLORREF bgcolor = GetRCColor("mkBackColor", 0x00000000);
		bgbrush = CreateSolidBrush(bgcolor);
	}

	//Default group = "Default" if not defined by user
	GetRCString("mkDefaultGroup", default_group, "Default", 256);
	default_group_defined = (GetRCBool("mkDefaultGroup", true) ? true:false);

	mkGroupShow = (GetRCBool("mkGroupShow", true) ? true:false);
	mkGroupX = GetRCInt("mkGroupX", 0);
	mkGroupY = GetRCInt("mkGroupY", 0);
	mkCommandShow = (GetRCBool("mkCommandShow", true) ? true:false);
	mkCommandX = GetRCInt("mkCommandX", 0);
	mkCommandY = GetRCInt("mkCommandY", 0);

	GetRCString("mkMsgErrorCmd", mkMsgErrorCmd, "Error: Command failed", 256);
	GetRCString("mkMsgErrorBang", mkMsgErrorBang, "Error: !Bang failed", 256);
	GetRCString("mkErrorSound", mkErrorSound, "", 256);

	// Assign commands to the mouse buttons
	mkMouseLeft = GetRCInt("mkMouseLeft", 1);
	mkMouseRight = GetRCInt("mkMouseRight", 0);
	mkMouseMiddle = GetRCInt("mkMouseMiddle", 2);

	// For future support of extended keyboards (such as Microsoft's Natural Keyboard)
	// mkExtended = (GetRCBool("mkExtended", true) ? true:false);

//--------------------

	// Create popup menus
	popup = CreatePopupMenu();
	popupgroups = CreatePopupMenu();

//--------------------

	// Look through step.rc for key config lines and set up hotkeys, popup menu items etc.

	FILE* step;
	char	token1[4096], token2[4096], token3[4096], token4[4096], token5[4096], arguments[4096];
	char*	tokens[5];
	tokens[0] = token1;
	tokens[1] = token2;
	tokens[2] = token3;
	tokens[3] = token4;
	tokens[4] = token5;

	// First we check for *mkey config lines...
	step = LCOpen(NULL);
	while (LCReadNextConfig(step, "*mkey", temp, 256)) 
	{ 
		int count;
		token1[0] = token2[0] = arguments[0] = '\0';
		count = LCTokenize (temp, tokens, 5, arguments);

		if (count == 5)
		{
			multikeysParseKey(token2, token3, token4, token5, arguments);
		}
	}
	LCClose(step);

	// Does the user want to disable loading of standard LS hotkey config lines?
	mkNoLoadHotkeys = (GetRCBool("mkNoLoadHotkeys", true) ? true:false);
	if (!mkNoLoadHotkeys) 
	{
		step = LCOpen(NULL);
		while (LCReadNextConfig(step, "*Hotkey", temp, 256)) 
		{ 
			int count;
			token1[0] = token2[0] = arguments[0] = '\0';
			count = LCTokenize (temp, tokens, 4, arguments);

			if (count == 4)
			{
				sprintf(token5, "%s", token4);
				sprintf(token4, "%s", token3);
				sprintf(token3, "%s", token2);
				sprintf(token2, "Hotkey");
				multikeysParseKey(token2, token3, token4, token5, arguments);
			}
		}
		LCClose(step);
	}

	// Does the user want to disable loading of jKey config lines?
	mkNoLoadjKeys = (GetRCBool("mkNoLoadjKeys", true) ? true:false);
	if (!mkNoLoadjKeys) 
	{
		step = LCOpen(NULL);
		while (LCReadNextConfig(step, "*jKey", temp, 256)) 
		{ 
			int count;
			token1[0] = token2[0] = arguments[0] = '\0';
			count = LCTokenize (temp, tokens, 4, arguments);

			if (count == 4)
			{
				sprintf(token5, "%s", token4);
				sprintf(token4, "%s", token3);
				sprintf(token3, "%s", token2);
				sprintf(token2, "jKey");
				multikeysParseKey(token2, token3, token4, token5, arguments);
			}
		}
		LCClose(step);
	}

	if (!default_group_defined)
	{
		for (Group* g = groups; g; g=g->next)
		{
			// If default group is not defined and Hotkeys exist, use "Hotkey" as default group
			// If default group is not defined and jKeys exist, use "jKey" as default group
			// jKey has precedence over Hotkey
			if (!_strcmpi("Hotkey", g->name))
			{
				sprintf(default_group, "Hotkey");
			}
			if (!_strcmpi("jKey", g->name))
			{
				sprintf(default_group, "jKey");
				break;
			}
		}
	}
}

//===========================================================================

void multikeysParseKey(char token2[4096], char token3[4096], char token4[4096], char token5[4096], char arguments[4096])
{
	Hotkey* mk = new Hotkey;
	char atomname[256] = "";
	
	sprintf(atomname, "mkey%d", id++);
	mk->ID = GlobalAddAtom(atomname);
	mk->group = _strdup(token2);
	mk->mods = 0;
			
	if (strlen(token4) == 1) mk->key = token4[0];
	else if (!_strcmpi(token4, "F1")) mk->key = VK_F1;
	else if (!_strcmpi(token4, "F2")) mk->key = VK_F2;
	else if (!_strcmpi(token4, "F3")) mk->key = VK_F3;
	else if (!_strcmpi(token4, "F4")) mk->key = VK_F4;
	else if (!_strcmpi(token4, "F5")) mk->key = VK_F5;
	else if (!_strcmpi(token4, "F6")) mk->key = VK_F6;
	else if (!_strcmpi(token4, "F7")) mk->key = VK_F7;
	else if (!_strcmpi(token4, "F8")) mk->key = VK_F8;
	else if (!_strcmpi(token4, "F9")) mk->key = VK_F9;
	else if (!_strcmpi(token4, "F10")) mk->key = VK_F10;
	else if (!_strcmpi(token4, "F11")) mk->key = VK_F11;
	else if (!_strcmpi(token4, "F12")) mk->key = VK_F12;
	else if (!_strcmpi(token4, "PrtScn")) mk->key = VK_SNAPSHOT;
	else if (!_strcmpi(token4, "Pause")) mk->key = VK_PAUSE;
	else if (!_strcmpi(token4, "Insert")) mk->key = VK_INSERT;
	else if (!_strcmpi(token4, "Delete")) mk->key = VK_DELETE;
	else if (!_strcmpi(token4, "Home")) mk->key = VK_HOME;
	else if (!_strcmpi(token4, "End")) mk->key = VK_END;
	else if (!_strcmpi(token4, "PageUp")) mk->key = VK_PRIOR;
	else if (!_strcmpi(token4, "PageDown")) mk->key = VK_NEXT;
	else if (!_strcmpi(token4, "Left")) mk->key = VK_LEFT;
	else if (!_strcmpi(token4, "Right")) mk->key = VK_RIGHT;
	else if (!_strcmpi(token4, "Down")) mk->key = VK_DOWN;
	else if (!_strcmpi(token4, "Up")) mk->key = VK_UP;
	else if (!_strcmpi(token4, "Tab")) mk->key = VK_TAB;
	else if (!_strcmpi(token4, "Backspace")) mk->key = VK_BACK;
	else if (!_strcmpi(token4, "Spacebar")) mk->key = VK_SPACE;
	else if (!_strcmpi(token4, "Apps")) mk->key = VK_APPS;
	else if (!_strcmpi(token4, "Enter")) mk->key = VK_RETURN;
	else if (!_strcmpi(token4, "Num0")) mk->key = VK_NUMPAD0;
	else if (!_strcmpi(token4, "Num1")) mk->key = VK_NUMPAD1;
	else if (!_strcmpi(token4, "Num2")) mk->key = VK_NUMPAD2;
	else if (!_strcmpi(token4, "Num3")) mk->key = VK_NUMPAD3;
	else if (!_strcmpi(token4, "Num4")) mk->key = VK_NUMPAD4;
	else if (!_strcmpi(token4, "Num5")) mk->key = VK_NUMPAD5;
	else if (!_strcmpi(token4, "Num6")) mk->key = VK_NUMPAD6;
	else if (!_strcmpi(token4, "Num7")) mk->key = VK_NUMPAD7;
	else if (!_strcmpi(token4, "Num8")) mk->key = VK_NUMPAD8;
	else if (!_strcmpi(token4, "Num9")) mk->key = VK_NUMPAD9;
	else if (!_strcmpi(token4, "Mul")) mk->key = VK_MULTIPLY;
	else if (!_strcmpi(token4, "Div")) mk->key = VK_DIVIDE;
	else if (!_strcmpi(token4, "Add")) mk->key = VK_ADD;
	else if (!_strcmpi(token4, "Sub")) mk->key = VK_SUBTRACT;
	else if (!_strcmpi(token4, "Dec")) mk->key = VK_DECIMAL;
	else if (!_strcmpi(token4, "Escape")) mk->key = VK_ESCAPE;

	// Extra keys to check if extended keyboard support is enabled
/*	if (mkExtended)
	{
		if (!_strcmpi(token4, "xBack")) mk->key = VK_BROWSER_BACK;
		else if (!_strcmpi(token4, "xForward")) mk->key = VK_BROWSER_FORWARD;
		else if (!_strcmpi(token4, "xStop")) mk->key = VK_BROWSER_STOP;
		else if (!_strcmpi(token4, "xRefresh")) mk->key = VK_BROWSER_REFRESH;
		else if (!_strcmpi(token4, "xSearch")) mk->key = VK_BROWSER_SEARCH;
		else if (!_strcmpi(token4, "xFav")) mk->key = VK_BROWSER_FAVORITES;
		else if (!_strcmpi(token4, "xWebHome")) mk->key = VK_BROWSER_HOME;
		else if (!_strcmpi(token4, "xMail")) mk->key = VK_LAUNCH_MAIL;
		else if (!_strcmpi(token4, "xMute")) mk->key = VK_VOLUME_MUTE;
		else if (!_strcmpi(token4, "xVolUp")) mk->key = VK_VOLUME_DOWN;
		else if (!_strcmpi(token4, "xVolDown")) mk->key = VK_VOLUME_UP;
		else if (!_strcmpi(token4, "xPlayPause")) mk->key = VK_MEDIA_PLAY_PAUSE;
		else if (!_strcmpi(token4, "xStop")) mk->key = VK_MEDIA_STOP;
		else if (!_strcmpi(token4, "xPrev")) mk->key = VK_MEDIA_PREV_TRACK;
		else if (!_strcmpi(token4, "xNext")) mk->key = VK_MEDIA_NEXT_TRACK;
		else if (!_strcmpi(token4, "xMedia")) mk->key = VK_LAUNCH_MEDIA_SELECT;
		else if (!_strcmpi(token4, "xApp1")) mk->key = VK_LAUNCH_APP1;
		else if (!_strcmpi(token4, "xApp2")) mk->key = VK_LAUNCH_APP2;
		else if (!_strcmpi(token4, "xSleep")) mk->key = VK_SLEEP;
	}
*/

	else { delete mk; return; } // FIX THIS!

	mk->cmd = _strdup(token5);
	mk->args = _strdup(arguments);

	char*	token3copy;
	token3copy = _strdup(token3);

	char* p = strtok(token3copy, "+");
	while (p)
	{
		if (!_strcmpi(p, "Win")) mk->mods |= MOD_WIN;
		if (!_strcmpi(p, "Alt")) mk->mods |= MOD_ALT;
		if (!_strcmpi(p, "Ctrl")) mk->mods |= MOD_CONTROL;
		if (!_strcmpi(p, "Shift")) mk->mods |= MOD_SHIFT;
				
	p = strtok(NULL, "+");
	}
		
	bool FOUND=false;
			
	if (groups)
	{
		for (Group* g = groups; g; g=g->next)
		{
			// Assign key to group if the group already exists
			if (!_strcmpi(mk->group, g->name))
			{
				FOUND=true;
				sprintf(temp, "%s %s %s %s", token3, token4, token5, arguments);
				AppendMenu(g->menu, MF_ENABLED | MF_STRING, mk->ID, temp);
			}
		}
	}

	if (!FOUND) 
	{
		// Add new group to "Groups" popup menu, and assign an unlikely hotkey modifier to it (starting with gKey -> "A")
		GroupHotkey* mkg = new GroupHotkey;
		sprintf(atomname, "mkeygroup%d", idgroup++);
		mkg->ID = GlobalAddAtom(atomname);
		mkg->cmd = "!mkGroup";
		mkg->args = _strdup(token2);

		// The unlikely modifier: Win+Alt+Control+Shift...
		mkg->mods = 0;
		mkg->mods |= MOD_WIN;
		mkg->mods |= MOD_ALT;
		mkg->mods |= MOD_CONTROL;
		mkg->mods |= MOD_SHIFT;

		// Assign a key (starting with A) to the group (-> starting with Win+Alt+Control+Shift A)
		mkg->key = gkey[0]++;

		AppendMenu(popupgroups, MF_ENABLED | MF_STRING, mkg->ID, token2);
		if (ghead) mkg->next = ghead;
		ghead = mkg;

		// Create a submenu for the new group in the main popup and assign key to that
		Group* g = new Group;
		g->menu = CreatePopupMenu();
		g->name = _strdup(mk->group);
		if (groups) g->next = groups;
		groups = g;
		AppendMenu(popup, MF_ENABLED | MF_STRING | MF_POPUP, (int)g->menu, mk->group);
		sprintf(temp, "%s %s %s %s", token3, token4, token5, arguments);
		AppendMenu(g->menu, MF_ENABLED | MF_STRING, mk->ID, temp);
	}

	if (head) mk->next = head;
	head = mk;
}

//===========================================================================
// Unregister hotkeys and destroy popup (used by quitmodule and !refresh)
//===========================================================================

void multikeysDestroyHotkeys()
{

	// Destroy popup menus + submenus
	DestroyMenu(popup);
	DestroyMenu(popupgroups);
	while (groups)
	{
		Group* g = groups;
		groups = groups->next;
		DestroyMenu(g->menu);
		delete g;
	}

	// Unregister hotkeys, free linked list
	while (head)
	{
		Hotkey* p = head;
		head = head->next;
		UnregisterHotKey(hwndMain, p->ID);
		GlobalDeleteAtom(p->ID);
		delete p;
	}

	// Unregister group hotkeys, free linked list
	while (ghead)
	{
		GroupHotkey* pg = ghead;
		ghead = ghead->next;
		UnregisterHotKey(hwndMain, pg->ID);
		GlobalDeleteAtom(pg->ID);
		delete pg;
	}
}

//===========================================================================
// Switch between groups of hotkeys
//===========================================================================

void LoadGroupHotkeys()
{
	// Register an "unlikely hotkey" for each group... (see multikeysLoadSetup)
	GroupHotkey* p;
	for (p=ghead; p; p=p->next)
	{
		RegisterHotKey(hwndMain, p->ID, p->mods, (int)p->key);
	}
}

void SwitchToGroup(const char* group)
{
	Hotkey* p;

	if (GroupExists(group)) {
		// Unregister everything first to make sure we have a clean start
		for (p=head; p; p=p->next)
		{
			if (!_strcmpi(p->group, current_group))
				UnregisterHotKey(hwndMain, p->ID);
		}

		// Now register the current group
		for (p=head; p; p=p->next)
		{
			if (!_strcmpi(p->group, group) || !_strcmpi(p->group, "All")) 
				RegisterHotKey(hwndMain, p->ID, p->mods, (int)p->key);
		}

		strcpy(current_group, group);
		InvalidateRect(hwndMain, NULL, true);
	}
}


bool GroupExists(const char* group)
{
	// Check if hotkey group exists
	for (Hotkey* p=head; p; p=p->next)
	{
		if (!_strcmpi(p->group, group)) return true;
	}

	return false;
}

//===========================================================================
// Code for !bangs and other commands
//===========================================================================

void multikeysDefaultGroup()
{
	SwitchToGroup(default_group);
}

//--------------------

void multikeysPopupMenu()
{
	DWORD dwp = GetMessagePos();
	TrackPopupMenu(popup, TPM_RIGHTALIGN | TPM_RIGHTBUTTON, LOWORD(dwp), HIWORD(dwp), 0, hwndMain, NULL);
}

void multikeysPopupGroups()
{
	DWORD dwp = GetMessagePos();
	TrackPopupMenu(popupgroups, TPM_RIGHTALIGN | TPM_RIGHTBUTTON, LOWORD(dwp), HIWORD(dwp), 0, hwndMain, NULL);
}

//--------------------

void multikeysRepeatCommand()
{
	if (last_cmd != "")
	{	
		LSExecuteEx(GetDesktopWindow(), NULL, last_cmd, last_args, NULL, SW_SHOWNORMAL);
	}
}

//--------------------

void multikeysSetCommand(HWND caller, const char *args)
{
	// Divide received data into command and arguments, copy to buffers, redraw
	if(args && *args)
	{
		char *temp, *newcmd, *newargs;
		temp = (char *)malloc(strlen(args) + 1);
		strcpy(temp, args);
		newcmd = strtok(temp, " \t");
		newargs = strtok(NULL, "");
		if(newcmd) sprintf(last_cmd, "%s", newcmd); else sprintf(last_cmd, "");
		if(newargs) sprintf(last_args, "%s", newargs); else sprintf(last_args, "");
		sprintf(LAST_COMMAND, "%s %s", newcmd, newargs);
		free(temp);
		InvalidateRect(hwndMain, NULL, true);
	}
}

void multikeysSetError(HWND caller, const char *args)
{
	// Copy error text to buffer, clear command repeat buffer, redraw
	if(args && *args)
	{
		error = true;
		sprintf(LAST_COMMAND, "%s", args);
		sprintf(last_cmd, "");
		sprintf(last_args, "");
		InvalidateRect(hwndMain, NULL, true);
	}
}

//--------------------

void multikeysRefreshPosition()
{
	POINT coord;

	ScreenX = GetSystemMetrics(SM_CXSCREEN);
	ScreenY = GetSystemMetrics(SM_CYSCREEN);
			
	x = rcx;
	y = rcy;
	multikeysGetPosition();
	coord.x = x;
	coord.y = y;

	SetWindowPos(hwndMain, NULL, coord.x, coord.y, 0, 0, SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
}

//--------------------

void multikeysMove(HWND caller, const char *args)
{
	POINT coord;

	ScreenX = GetSystemMetrics(SM_CXSCREEN);
	ScreenY = GetSystemMetrics(SM_CYSCREEN);

	// Get original step.rc position (if no arguments were used we move to this position)
	x = rcx;
	y = rcy;
	multikeysGetPosition();
	coord.x = x;
	coord.y = y;

	// Are we moving to a new position? Did we get both x and y arguments?
	if(args && *args)
	{
		char *temp, *newx, *newy;
		temp = (char *)malloc(strlen(args) + 1);
		strcpy(temp, args);
		newx = strtok(temp, " \t");
		newy = strtok(NULL, "");
		if(newx) coord.x = atoi(newx);
		if(newy) coord.y = atoi(newy); else { coord.x = x; coord.y = y; }
		free(temp);
	}

	//MoveWindow(hwndMain, coord.x, coord.y, w, h, true);
	SetWindowPos(hwndMain, NULL, coord.x, coord.y, 0, 0, SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
}

//===========================================================================
// Calculate screen coordinates
//===========================================================================

void multikeysGetPosition()
{
	if (mkOffsetX == 0) { if (x < 0) x = ScreenX + x; }
	else if (mkOffsetX == 1) { x = ScreenX/2 - w/2 + x;}
	else if (mkOffsetX == 2) { x = ScreenX + x; }

	if (mkOffsetY == 0) { if (y < 0) y = ScreenY + y; }
	else if (mkOffsetY == 1) { y = ScreenY/2 - h/2 + y; }
	else if (mkOffsetY == 2) { y = ScreenY + y; }
}

//===========================================================================
// SetWindowBitmapRgn
//===========================================================================

void SetWindowBitmapRgn(HWND hwnd, HBITMAP bmp)
{
	if (hwnd && bmp)
	{
		int x=0, y=0;
		HDC hdc = CreateCompatibleDC(NULL);
		HRGN hTransRgn=NULL;
		HRGN hMainRgn=NULL;
		BITMAP bitmap;

		GetObject(bmp, sizeof(bitmap), &bitmap);
		SelectObject(hdc, bmp);
		hMainRgn = CreateRectRgn(0, 0, bitmap.bmWidth, bitmap.bmHeight);

		for (y=0; y < bitmap.bmHeight; x=0, y++)
		{
			for (x=0; x < bitmap.bmWidth; x++)
			{
				COLORREF c = GetPixel(hdc, x, y);
				if (c == 0x00FF00FF)
				{
					HRGN hTempRgn = CreateRectRgn(x, y, x+1, y+1);
					if (!hTransRgn) hTransRgn = CreateRectRgn(x, y, x+1, y+1);
					CombineRgn(hTransRgn, hTransRgn, hTempRgn, RGN_OR);
					DeleteObject(hTempRgn);
				}
			}
		}
		
		CombineRgn(hMainRgn, hMainRgn, hTransRgn, RGN_DIFF);
		SetWindowRgn(hwnd, hMainRgn, false);

		DeleteDC(hdc);
		DeleteObject(hTransRgn);
		DeleteObject(hMainRgn);
	}
	else
	{
		RECT r;
		HRGN rgn;
		GetClientRect(hwnd, &r);
		rgn = CreateRectRgn(r.left, r.top, r.right, r.bottom);
		SetWindowRgn(hwnd, rgn, false);
	}
}

//===========================================================================
// Message handling
//===========================================================================

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{

//--------------------

		case WM_LBUTTONDOWN:
		{
			if (mkMouseLeft == 0) multikeysPopupMenu();
			else if (mkMouseLeft == 1) multikeysPopupGroups();
			else if (mkMouseLeft == 2) multikeysRepeatCommand();
			else if (mkMouseLeft == 3) multikeysDefaultGroup();
		}
		return 0;

		case WM_LBUTTONUP: { } break;
		
		case WM_RBUTTONDOWN:
		{
			if (mkMouseRight == 0) multikeysPopupMenu();
			else if (mkMouseRight == 1) multikeysDefaultGroup();
			else if (mkMouseRight == 2) multikeysRepeatCommand();
			else if (mkMouseRight == 3) multikeysDefaultGroup();
		}
		return 0;

		case WM_RBUTTONUP: { } break;

		case WM_MBUTTONDOWN:
		{
			if (mkMouseMiddle == 0) multikeysPopupMenu();
			else if (mkMouseMiddle == 1) multikeysDefaultGroup();
			else if (mkMouseMiddle == 2) multikeysRepeatCommand();
			else if (mkMouseMiddle == 3) multikeysDefaultGroup();
		}
		return 0;

		case WM_MBUTTONUP: { } break;
	
//--------------------
	
		case WM_ERASEBKGND: return 0;

//--------------------


		case LM_GETREVID:
		{
			char *buf = (char *) lParam;

			if (wParam == 0)
			{
				strcpy(buf, szVersion);
			}
			else if (wParam == 1)
			{
				strcpy(buf, szDetailed);
			} else
			{
				strcpy(buf, "");
			}
			return strlen(buf);
		}
		break;

		case LM_REFRESH:
		{
			multikeysDestroyHotkeys();
			multikeysLoadSetup();
			multikeysRefreshPosition();
			sprintf(LAST_COMMAND, "!Refresh");
			sprintf(last_cmd, "!Refresh");
			sprintf(last_args, "");
			SwitchToGroup(default_group);
		}
		break;

//--------------------

		case WM_DISPLAYCHANGE:
		{
			multikeysRefreshPosition();
		}
		break;

//--------------------

		case WM_COMMAND:
		case WM_HOTKEY:
		{
			// Look for matching hotkey
			for (Hotkey* p=head; p; p=p->next)
			{
				if ((int)wParam == p->ID)
				{
					if (((int)LSExecuteEx(GetDesktopWindow(), NULL, p->cmd, p->args, NULL, SW_SHOWNORMAL)) <= 32)
					{
						error = true;
						if (p->cmd[0] == '!') {	sprintf(LAST_COMMAND, mkMsgErrorBang); }
						else { sprintf(LAST_COMMAND, mkMsgErrorCmd); }
						sprintf(last_cmd, "");
						sprintf(last_args, "");
						InvalidateRect(hwndMain, NULL, true);
						PlaySound(mkErrorSound, NULL, SND_FILENAME | SND_ASYNC | SND_NOWAIT);
					}
					else
					{
						error = false;
						sprintf(LAST_COMMAND, "%s %s", p->cmd, p->args);
						sprintf(last_cmd, "%s", p->cmd);
						sprintf(last_args, "%s", p->args);
						InvalidateRect(hwndMain, NULL, true);
					}
					break;
				}
			}

			// Look for matching group hotkey
			for (GroupHotkey* pg=ghead; pg; pg=pg->next)
			{
				if ((int)wParam == pg->ID)
				{
					LSExecuteEx(GetDesktopWindow(), NULL, pg->cmd, pg->args, NULL, SW_SHOWNORMAL);
					break;
				}
			}

		}
		break;

//--------------------

		case WM_PAINT:
		{
			PAINTSTRUCT ps;
			HDC hdc = BeginPaint(hwnd, &ps);
			HDC buf = CreateCompatibleDC(NULL);
			HDC src = CreateCompatibleDC(NULL);
			HBITMAP bufbmp = CreateCompatibleBitmap(hdc, w, h);
			HBITMAP oldbuf, oldsrc;
			RECT r;

			GetClientRect(hwnd, &r);

			if (!bgbitmap)
			{
				bgbitmap = CreateCompatibleBitmap(hdc, w, h);
				oldbuf = (HBITMAP)SelectObject(buf, bgbitmap);
				BitBlt(buf, 0, 0, w, h, hdc, 0, 0, SRCCOPY);
				SelectObject(buf, oldbuf);
			}

			oldbuf = (HBITMAP)SelectObject(buf, bufbmp);
			
			if (UseBgColor && !UseBgBitmap)
			{
				FillRect(buf, &r, (HBRUSH)bgbrush);
			}
			else
			{
				oldsrc = (HBITMAP)SelectObject(src, bgbitmap);
				BitBlt(buf, 0, 0, w, h, src, 0, 0, SRCCOPY);
			}

			if (mkGroupShow)
			{
				SelectObject(buf, font);
				SetBkMode(buf, TRANSPARENT);
				SetTextColor(buf, color);
				r.left = mkGroupX;
				r.top = mkGroupY;
				DrawText(buf, current_group, strlen(current_group), &r, DT_LEFT | DT_TOP | DT_WORD_ELLIPSIS);
				r.left = r.top = 0;
			}

			if (mkCommandShow && !error)
			{
				SelectObject(buf, font);
				SetBkMode(buf, TRANSPARENT);
				SetTextColor(buf, cmdcolor);
				r.left = mkCommandX;
				r.top = mkCommandY;
				r.right = w - mkGroupX;
				DrawText(buf, LAST_COMMAND, strlen(LAST_COMMAND), &r, DT_LEFT | DT_TOP | DT_WORD_ELLIPSIS);
				r.left = r.top = r.right = 0;
			}

			if (mkCommandShow && error)
			{
				SelectObject(buf, font);
				SetBkMode(buf, TRANSPARENT);
				SetTextColor(buf, errorcolor);
				r.left = mkCommandX;
				r.top = mkCommandY;
				r.right = w - mkGroupX;
				DrawText(buf, LAST_COMMAND, strlen(LAST_COMMAND), &r, DT_LEFT | DT_TOP | DT_WORD_ELLIPSIS);
				r.left = r.top = r.right = 0;
				error = false;
			}
			else { error = false; }

			BitBlt(hdc, 0, 0, w, h, buf, 0, 0, SRCCOPY);

			SelectObject(src, oldsrc);
			DeleteDC(src);
			SelectObject(buf, oldbuf);
			DeleteDC(buf);
			DeleteObject(bufbmp);
			EndPaint(hwnd, &ps);
		}
		break;

//--------------------

		case WM_DESTROY:
		{
			if (HookedToLsBox)
			{
				SetWindowLong(hwndMain, GWL_STYLE, (GetWindowLong(hwndMain, GWL_STYLE) &~ WS_CHILD)|WS_POPUP);
				SetParent(hwndMain, 0);
				SetWindowPos(hwndMain,HWND_TOP,x,y,0,0,SWP_NOSIZE|SWP_HIDEWINDOW);
				HookedToLsBox = false;
				return 0;
			}
		}
		break;

//--------------------

	}

	return DefWindowProc(hwnd,message, wParam,lParam);
}

//===========================================================================
