/*
 ============================================================================
 multikeys - an advanced hotkeys module for LiteStep/PureLS
 ============================================================================

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

 ============================================================================ */

#ifndef __MULTIKEYS_H
#define __MULTIKEYS_H

#include <windows.h>
#include <mmsystem.h>
#include "../lsapi/lsapi.h"
//#include "exports.h"

//--------------------

LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
HINSTANCE hInstance;
HWND hwndMain, hwndParent;
int msgs[] = {LM_GETREVID, LM_REFRESH, 0};

//#define magicDWord 0x49474541

//===========================================================================
// Declarations
//===========================================================================

void BangShow(HWND caller, const char* args);
void BangHide(HWND caller, const char* args);
void BangToggle(HWND caller, const char* args);
void BangPopup(HWND caller, const char* args);
void BangPopupGroups(HWND caller, const char* args);
void BangGroup(HWND caller, const char* args);
void BangRepeat(HWND caller, const char* args);
void BangMove(HWND caller, const char *args);
void BangSetCommand(HWND caller, const char *args);
void BangSetError(HWND caller, const char *args);
void BangPlayWav(HWND caller, const char *args);

void multikeysAddBangs();
void multikeysRemoveBangs();
void multikeysLoadSetup();
void multikeysParseKey(char token2[4096], char token3[4096], char token4[4096], char token5[4096], char extra_text[4096]);
void multikeysDestroyHotkeys();

void LoadGroupHotkeys();
void SwitchToGroup(const char* group);
bool GroupExists(const char* group);

void multikeysDefaultGroup();
void multikeysPopupMenu();
void multikeysPopupGroups();
void multikeysRepeatCommand();
void multikeysSetCommand(HWND caller, const char *args);
void multikeysSetError(HWND caller, const char *args);
void multikeysRefreshPosition();
void multikeysMove(HWND caller, const char *args);

void multikeysGetPosition();

void SetWindowBitmapRgn(HWND hwnd, HBITMAP bmp);

//===========================================================================
// Variables
//===========================================================================

int ScreenX, ScreenY;
int mkOffsetX, mkOffsetY;
int x, y, w, h;
int rcx, rcy;			//original step.rc x y position ("backup copy"...)

bool mkStartHidden, mkLsBoxEnable;
bool HookedToLsBox;

bool mkGroupShow, mkCommandShow;
int mkGroupX, mkGroupY, mkCommandX, mkCommandY;

int mkMouseLeft, mkMouseRight, mkMouseMiddle;

bool UseBgColor;
bool UseBgBitmap;
bool mkAlwaysOnTop;
bool mkExtended;
bool error = false;
bool mkNoLoadHotkeys = false;
bool mkNoLoadjKeys = false;

HMENU popup=NULL, popupgroups=NULL;
HFONT font;
COLORREF color, cmdcolor, errorcolor;
HBRUSH bgbrush;
HBITMAP bgbitmap;

char temp[256];
char LSdir[256] = "";
char current_group[256] = "";
char default_group[256] = "";
bool default_group_defined;
char LAST_COMMAND[256] = "";
char last_cmd[256] = "";
char last_args[256] = "";
char mkMsgErrorCmd[256] = "";
char mkMsgErrorBang[256] = "";
char mkErrorSound[256] = "";
char mkPlayWav[256] = "";

int id=0, idgroup=0;
char gkey[2] = "A";

//===========================================================================
// Hotkey & group classes
//===========================================================================

class Group
{
public:
	Group() { menu=NULL; next=NULL; }
	char* name;
	HMENU menu;
	Group* next;
} *groups;

class Hotkey
{
public:
	Hotkey() { memset(&cmd, 0, sizeof(cmd)); next=NULL; }
	int ID;
	char* group;
	char* cmd;
	char* args;
	DWORD mods;
	char key;
	Hotkey* next;
} *head;

class GroupHotkey
{
public:
	GroupHotkey() { memset(&cmd, 0, sizeof(cmd)); next=NULL; }
	int ID;
	char* cmd;
	char* args;
	DWORD mods;
	char key;
	GroupHotkey* next;
} *ghead;

//===========================================================================
// External
//===========================================================================

extern "C"
{
	__declspec( dllexport ) int initModuleEx(HWND parent, HINSTANCE dllInst, LPCSTR szPath);
	__declspec( dllexport ) void quitModule(HINSTANCE dll);
}

#endif
